from __future__ import annotations

import sys
import json
import asyncio

import zigpy_znp.types as t
import zigpy_znp.config as conf
from zigpy_znp.types.nvids import OsalNvIds
from zigpy_znp.tools.common import ClosableFileType, setup_parser, validate_backup_json
from zigpy_znp.znp.security import StoredDevice, write_devices, write_tc_frame_counter
from zigpy_znp.zigbee.application import ControllerApplication


async def restore_network(
    radio_path: str,
    backup: t.JSONType,
    counter_increment: int,
) -> None:
    pan_id, _ = t.NWK.deserialize(bytes.fromhex(backup["pan_id"])[::-1])
    extended_pan_id, _ = t.EUI64.deserialize(
        bytes.fromhex(backup["extended_pan_id"])[::-1]
    )
    coordinator_ieee, _ = t.EUI64.deserialize(
        bytes.fromhex(backup["coordinator_ieee"])[::-1]
    )
    nwk_key, _ = t.KeyData.deserialize(bytes.fromhex(backup["network_key"]["key"]))

    devices = []

    for obj in backup["devices"]:
        nwk, _ = t.NWK.deserialize(bytes.fromhex(obj["nwk_address"])[::-1])
        ieee, _ = t.EUI64.deserialize(bytes.fromhex(obj["ieee_address"])[::-1])

        device = StoredDevice(nwk=nwk, ieee=ieee)

        if "link_key" in obj:
            key, _ = t.KeyData.deserialize(bytes.fromhex(obj["link_key"]["key"]))
            device = device.replace(
                aps_link_key=key,
                tx_counter=obj["link_key"]["tx_counter"],
                rx_counter=obj["link_key"]["rx_counter"],
            )

        devices.append(device)

    config = ControllerApplication.SCHEMA({"device": {"path": radio_path}})
    app = ControllerApplication(config)

    app.config[conf.CONF_NWK][conf.CONF_NWK_KEY] = nwk_key
    app.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID] = pan_id
    app.config[conf.CONF_NWK][conf.CONF_NWK_CHANNEL] = backup["channel"]
    app.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID] = extended_pan_id

    await app.startup(force_form=True)

    znp = app._znp

    await znp.load_network_info()
    await znp.reset()

    await znp.nvram.osal_write(OsalNvIds.EXTADDR, coordinator_ieee)

    nib = await znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
    nib.channelList = t.Channels.from_channel_list(backup["channel_mask"])
    nib.nwkUpdateId = backup["nwk_update_id"]
    nib.SecurityLevel = backup["security_level"]
    await znp.nvram.osal_write(OsalNvIds.NIB, nib)

    tclk_seed = None

    if znp.version > 1.20:
        if backup.get("stack_specific", {}).get("zstack", {}).get("tclk_seed"):
            tclk_seed = bytes.fromhex(backup["stack_specific"]["zstack"]["tclk_seed"])
            await znp.nvram.osal_write(OsalNvIds.TCLK_SEED, tclk_seed)

    nwk_frame_counter = backup["network_key"]["frame_counter"]
    nwk_frame_counter += counter_increment

    key_info = t.NwkActiveKeyItems(
        Active=t.NwkKeyDesc(
            KeySeqNum=backup["network_key"]["sequence_number"],
            Key=nwk_key,
        ),
        FrameCounter=nwk_frame_counter,
    )

    await znp.nvram.osal_write(OsalNvIds.NWKKEY, key_info)
    await znp.nvram.osal_write(OsalNvIds.NWK_ACTIVE_KEY_INFO, key_info.Active)
    await znp.nvram.osal_write(OsalNvIds.NWK_ALTERN_KEY_INFO, key_info.Active)
    await write_tc_frame_counter(znp, nwk_frame_counter)

    await write_devices(
        znp, devices, tclk_seed=tclk_seed, counter_increment=counter_increment
    )

    await znp.reset()

    znp.close()


async def main(argv: list[str]) -> None:
    parser = setup_parser("Restore adapter network settings")
    parser.add_argument(
        "--input", "-i", type=ClosableFileType("r"), help="Input file", required=True
    )
    parser.add_argument(
        "--counter-increment",
        "-c",
        type=t.uint32_t,
        help="Counter increment",
        default=2500,
    )
    args = parser.parse_args(argv)

    with args.input as f:
        backup = json.load(f)

    validate_backup_json(backup)

    await restore_network(
        radio_path=args.serial,
        backup=backup,
        counter_increment=args.counter_increment,
    )


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover



import dataclasses

import zigpy_znp.types as t
from zigpy_znp.exceptions import CommandNotRecognized
from zigpy_znp.types.nvids import ExNvIds, OsalNvIds


@dataclasses.dataclass(frozen=True)
class NetworkInfo:
    ieee: t.EUI64
    nwk: t.NWK
    channel: t.uint8_t
    channels: t.Channels
    pan_id: t.NWK
    extended_pan_id: t.EUI64
    nwk_update_id: t.uint8_t
    security_level: t.uint8_t
    network_key: t.KeyData
    network_key_seq: t.uint8_t

    def replace(self, **kwargs):
        return dataclasses.replace(self, **kwargs)


async def load_network_info(znp) -> NetworkInfo:
    """
    Loads low-level network information from NVRAM.
    """

    is_on_network = None
    nib = None

    try:
        nib = await znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
        is_on_network = nib.nwkLogicalChannel != 0 and nib.nwkKeyLoaded
    except KeyError:
        is_on_network = False
    else:
        if is_on_network and znp.version >= 3.0:
            # This NVRAM item is the very first thing initialized in `zgInit`, it exists
            is_on_network = (
                await znp.nvram.osal_read(
                    OsalNvIds.BDBNODEISONANETWORK, item_type=t.uint8_t
                )
                == 1
            )

    if not is_on_network:
        raise ValueError("Device is not a part of a network")

    ieee = await znp.nvram.osal_read(OsalNvIds.EXTADDR, item_type=t.EUI64)
    key_desc = await znp.nvram.osal_read(
        OsalNvIds.NWK_ACTIVE_KEY_INFO, item_type=t.NwkKeyDesc
    )

    return NetworkInfo(
        ieee=ieee,
        nwk=nib.nwkDevAddress,
        channel=nib.nwkLogicalChannel,
        channels=nib.channelList,
        pan_id=nib.nwkPanId,
        extended_pan_id=nib.extendedPANID,
        nwk_update_id=nib.nwkUpdateId,
        security_level=nib.SecurityLevel,
        network_key=key_desc.Key,
        network_key_seq=key_desc.KeySeqNum,
    )


async def detect_zstack_version(znp) -> float:
    """
    Feature detects the major version of Z-Stack running on the device.
    """

    # Z-Stack 1.2 does not have the AppConfig subsystem
    if not znp.capabilities & t.MTCapabilities.APP_CNF:
        return 1.2

    try:
        # Only Z-Stack 3.30+ has the new NVRAM system
        await znp.nvram.read(
            item_id=ExNvIds.TCLK_TABLE,
            sub_id=0x0000,
            item_type=t.Bytes,
        )
        return 3.30
    except KeyError:
        return 3.30
    except CommandNotRecognized:
        return 3.0



from __future__ import annotations

import time
import asyncio
import logging
import itertools
import contextlib
from collections import Counter, defaultdict

import async_timeout

import zigpy_znp.types as t
import zigpy_znp.config as conf
import zigpy_znp.logger as log
import zigpy_znp.commands as c
from zigpy_znp import uart
from zigpy_znp.nvram import NVRAMHelper
from zigpy_znp.utils import (
    BaseResponseListener,
    OneShotResponseListener,
    CallbackResponseListener,
)
from zigpy_znp.frames import GeneralFrame
from zigpy_znp.znp.utils import NetworkInfo, load_network_info, detect_zstack_version
from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse

LOGGER = logging.getLogger(__name__)
AFTER_CONNECT_DELAY = 1  # seconds
STARTUP_DELAY = 1  # seconds


class ZNP:
    def __init__(self, config: conf.ConfigType):
        self._uart = None
        self._app = None
        self._config = config

        self._listeners = defaultdict(list)
        self._sync_request_lock = asyncio.Lock()

        self.capabilities = None
        self.version = None

        self.nvram = NVRAMHelper(self)
        self.network_info: NetworkInfo = None

    def set_application(self, app):
        assert self._app is None
        self._app = app

    async def load_network_info(self):
        self.network_info = await load_network_info(self)

    async def reset(self) -> None:
        """
        Performs a soft reset within Z-Stack.
        A hard reset resets the serial port, causing the device to disconnect.
        """

        await self.request_callback_rsp(
            request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
            callback=c.SYS.ResetInd.Callback(partial=True),
        )

    @property
    def _port_path(self) -> str:
        return self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH]

    async def connect(self, *, test_port=True) -> None:
        """
        Connects to the device specified by the "device" section of the config dict.

        The `test_port` kwarg allows port testing to be disabled, mainly to get into the
        bootloader.
        """

        # So we cannot connect twice
        assert self._uart is None

        try:
            self._uart = await uart.connect(self._config[conf.CONF_DEVICE], self)

            LOGGER.debug("Waiting %ss before sending anything", AFTER_CONNECT_DELAY)
            await asyncio.sleep(AFTER_CONNECT_DELAY)

            if self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SKIP_BOOTLOADER]:
                LOGGER.debug("Sending bootloader skip byte")

                # XXX: Z-Stack locks up if other radios try probing it first.
                #      Writing the bootloader skip byte a bunch of times (at least 167)
                #      appears to reset it.
                skip = bytes([c.ubl.BootloaderRunMode.FORCE_RUN])
                self._uart._transport_write(skip * 256)

            # We have to disable all non-bootloader commands to enter the serial
            # bootloader upon connecting to the UART.
            if test_port:
                # Some Z-Stack 3 devices don't like you sending data immediately after
                # opening the serial port. A small delay helps, but they also sometimes
                # send a reset indication message when they're ready.
                LOGGER.debug(
                    "Waiting %ss or until a reset indication is received", STARTUP_DELAY
                )

                try:
                    async with async_timeout.timeout(STARTUP_DELAY):
                        await self.wait_for_response(
                            c.SYS.ResetInd.Callback(partial=True)
                        )
                except asyncio.TimeoutError:
                    pass

                LOGGER.debug("Testing connection to %s", self._port_path)

                # Make sure that our port works
                self.capabilities = (await self.request(c.SYS.Ping.Req())).Capabilities

                # We need to know how structs are packed to deserialize frames corectly
                await self.nvram.determine_alignment()
                self.version = await detect_zstack_version(self)

                LOGGER.debug("Detected Z-Stack %s", self.version)
        except (Exception, asyncio.CancelledError):
            LOGGER.debug("Connection to %s failed, cleaning up", self._port_path)
            self.close()
            raise

        LOGGER.debug(
            "Connected to %s at %s baud",
            self._uart._transport.serial.name,
            self._uart._transport.serial.baudrate,
        )

    def connection_made(self) -> None:
        """
        Called by the UART object when a connection has been made.
        """
        pass

    def connection_lost(self, exc) -> None:
        """
        Called by the UART object to indicate that the port was closed. Propagates up
        to the `ControllerApplication` that owns this ZNP instance.
        """

        LOGGER.debug("We were disconnected from %s: %s", self._port_path, exc)

        if self._app is not None:
            self._app.connection_lost(exc)

    def close(self) -> None:
        """
        Cleans up resources, namely the listener queues.

        Calling this will reset ZNP to the same internal state as a fresh ZNP instance.
        """

        self._app = None

        for header, listeners in self._listeners.items():
            for listener in listeners:
                listener.cancel()

        self._listeners.clear()
        self.version = None
        self.capabilities = None

        if self._uart is not None:
            self._uart.close()
            self._uart = None

    def remove_listener(self, listener: BaseResponseListener) -> None:
        """
        Unbinds a listener from ZNP.

        Used by `wait_for_responses` to remove listeners for completed futures,
        regardless of their completion reason.
        """

        # If ZNP is closed while it's still running, `self._listeners` will be empty.
        if not self._listeners:
            return

        LOGGER.log(log.TRACE, "Removing listener %s", listener)

        for header in listener.matching_headers():
            try:
                self._listeners[header].remove(listener)
            except ValueError:
                pass

            if not self._listeners[header]:
                LOGGER.log(
                    log.TRACE, "Cleaning up empty listener list for header %s", header
                )
                del self._listeners[header]

        counts = Counter()

        for listener in itertools.chain.from_iterable(self._listeners.values()):
            counts[type(listener)] += 1

        LOGGER.log(
            log.TRACE,
            "There are %d callbacks and %d one-shot listeners remaining",
            counts[CallbackResponseListener],
            counts[OneShotResponseListener],
        )

    def frame_received(self, frame: GeneralFrame) -> bool:
        """
        Called when a frame has been received. Returns whether or not the frame was
        handled by any listener.

        XXX: Can be called multiple times in a single event loop step!
        """

        if frame.header not in c.COMMANDS_BY_ID:
            LOGGER.error("Received an unknown frame: %s", frame)
            return

        command_cls = c.COMMANDS_BY_ID[frame.header]

        try:
            command = command_cls.from_frame(frame, align=self.nvram.align_structs)
        except ValueError:
            # Some commands can be received corrupted. They are not useful:
            # https://github.com/home-assistant/core/issues/50005
            if command_cls == c.ZDO.ParentAnnceRsp.Callback:
                LOGGER.warning("Failed to parse broken %s as %s", frame, command_cls)
                return

            raise

        LOGGER.debug("Received command: %s", command)

        matched = False
        one_shot_matched = False

        for listener in self._listeners[command.header]:
            # XXX: A single response should *not* resolve multiple one-shot listeners!
            #      `future.add_done_callback` doesn't remove our listeners synchronously
            #      so doesn't prevent this from happening.
            if one_shot_matched and isinstance(listener, OneShotResponseListener):
                continue

            if not listener.resolve(command):
                LOGGER.log(log.TRACE, "%s does not match %s", command, listener)
                continue

            matched = True
            LOGGER.log(log.TRACE, "%s matches %s", command, listener)

            if isinstance(listener, OneShotResponseListener):
                one_shot_matched = True

        if not matched:
            self._unhandled_command(command)

        return matched

    def _unhandled_command(self, command: t.CommandBase):
        """
        Called when a command that is not handled by any listener is received.
        """

        LOGGER.warning("Received an unhandled command: %s", command)

    @contextlib.asynccontextmanager
    async def capture_responses(self, responses):
        """
        Captures all matched responses in a queue within the context manager.
        """

        queue = asyncio.Queue()
        listener = self.callback_for_responses(responses, callback=queue.put_nowait)

        try:
            yield queue
        finally:
            self.remove_listener(listener)

    def callback_for_responses(self, responses, callback) -> CallbackResponseListener:
        """
        Creates a callback listener that matches any of the provided responses.

        Only exists for consistency with `wait_for_responses`, since callbacks can be
        executed more than once.
        """

        listener = CallbackResponseListener(responses, callback=callback)

        LOGGER.log(log.TRACE, "Creating callback %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        return listener

    def callback_for_response(
        self, response: t.CommandBase, callback
    ) -> CallbackResponseListener:
        """
        Creates a callback listener for a single response.
        """

        return self.callback_for_responses([response], callback)

    def wait_for_responses(self, responses, *, context=False) -> asyncio.Future:
        """
        Creates a one-shot listener that matches any *one* of the given responses.
        """

        listener = OneShotResponseListener(responses)

        LOGGER.log(log.TRACE, "Creating one-shot listener %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        # Remove the listener when the future is done, not only when it gets a result
        listener.future.add_done_callback(lambda _: self.remove_listener(listener))

        if context:
            return listener.future, listener
        else:
            return listener.future

    def wait_for_response(self, response: t.CommandBase) -> asyncio.Future:
        """
        Creates a one-shot listener for a single response.
        """

        return self.wait_for_responses([response])

    async def request(self, request: t.CommandBase, **response_params) -> t.CommandBase:
        """
        Sends a SREQ/AREQ request and returns its SRSP (only for SREQ), failing if any
        of the SRSP's parameters don't match `response_params`.
        """

        # Common mistake is to do `znp.request(c.SYS.Ping())`
        if type(request) is not request.Req:
            raise ValueError(f"Cannot send a command that isn't a request: {request!r}")

        # Construct a partial response out of the `Rsp*` kwargs if one is provided
        if request.Rsp:
            renamed_response_params = {}

            for param, value in response_params.items():
                if not param.startswith("Rsp"):
                    raise KeyError(
                        f"All response params must start with 'Rsp': {param!r}"
                    )

                renamed_response_params[param.replace("Rsp", "", 1)] = value

            # Construct our response before we send the request so that we fail early
            partial_response = request.Rsp(partial=True, **renamed_response_params)
        elif response_params:
            raise ValueError(
                f"Command has no response so response_params={response_params} "
                f"will have no effect"
            )

        frame = request.to_frame(align=self.nvram.align_structs)

        # We should only be sending one SREQ at a time, according to the spec
        async with self._sync_request_lock:
            LOGGER.debug("Sending request: %s", request)

            # If our request has no response, we cannot wait for one
            if not request.Rsp:
                LOGGER.debug("Request has no response, not waiting for one.")
                self._uart.send(frame)
                return

            # We need to create the response listener before we send the request
            response_future = self.wait_for_responses(
                [
                    request.Rsp(partial=True),
                    c.RPCError.CommandNotRecognized.Rsp(
                        partial=True, RequestHeader=request.header
                    ),
                ]
            )

            self._uart.send(frame)

            # We should get a SRSP in a reasonable amount of time
            async with async_timeout.timeout(
                self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SREQ_TIMEOUT]
            ):
                # We lock until either a sync response is seen or an error occurs
                response = await response_future

        if isinstance(response, c.RPCError.CommandNotRecognized.Rsp):
            raise CommandNotRecognized(
                f"Fatal request error {response} in response to {request}"
            )

        # If the sync response we got is not what we wanted, this is an error
        if not partial_response.matches(response):
            raise InvalidCommandResponse(
                f"Expected SRSP response {partial_response}, got {response}", response
            )

        return response

    async def request_callback_rsp(
        self, *, request, callback, timeout=None, background=False, **response_params
    ):
        """
        Sends an SREQ, gets its SRSP confirmation, and waits for its real AREQ response.
        A bug-free version of:

            req_rsp = await req
            callback_rsp = await req_callback

        This is necessary because the SRSP and the AREQ may arrive in the same "chunk"
        from the UART and be handled in the same event loop step by ZNP.
        """

        # Every request should have a timeout to prevent deadlocks
        if timeout is None:
            timeout = self._config[conf.CONF_ZNP_CONFIG][conf.CONF_ARSP_TIMEOUT]

        callback_rsp, listener = self.wait_for_responses([callback], context=True)

        if not background:
            try:
                async with async_timeout.timeout(timeout):
                    await self.request(request, **response_params)

                    return await callback_rsp
            finally:
                self.remove_listener(listener)

        start_time = time.time()

        # If the SREQ/SRSP pair fails, we must cancel the AREQ listener
        try:
            async with async_timeout.timeout(timeout):
                request_rsp = await self.request(request, **response_params)
        except Exception:
            self.remove_listener(listener)
            raise

        async def callback_handler(timeout):
            try:
                async with async_timeout.timeout(timeout):
                    await callback_rsp
            finally:
                self.remove_listener(listener)

        # If it succeeds, create a background task to receive the AREQ but take into
        # account the time it took to start the SREQ to ensure we do not grossly exceed
        # the timeout
        asyncio.create_task(callback_handler(time.time() - start_time))

        return request_rsp



from __future__ import annotations

import sys
import json
import asyncio
import logging
import datetime

import zigpy_znp
import zigpy_znp.types as t
from zigpy_znp.api import ZNP
from zigpy_znp.types.nvids import OsalNvIds
from zigpy_znp.tools.common import ClosableFileType, setup_parser, validate_backup_json
from zigpy_znp.znp.security import read_devices, read_tc_frame_counter
from zigpy_znp.zigbee.application import ControllerApplication

LOGGER = logging.getLogger(__name__)


async def backup_network(znp: ZNP) -> t.JSONType:
    try:
        await znp.load_network_info()
    except ValueError as e:
        raise RuntimeError("Failed to load network info") from e

    devices = []

    for device in await read_devices(znp):
        obj = {
            "nwk_address": device.nwk.serialize()[::-1].hex(),
            "ieee_address": device.ieee.serialize()[::-1].hex(),
        }

        if device.aps_link_key:
            obj["link_key"] = {
                "tx_counter": device.tx_counter,
                "rx_counter": device.rx_counter,
                "key": device.aps_link_key.serialize().hex(),
            }

        devices.append(obj)

    devices.sort(key=lambda d: d["ieee_address"])

    now = datetime.datetime.now().astimezone()

    obj = {
        "metadata": {
            "version": 1,
            "format": "zigpy/open-coordinator-backup",
            "source": f"zigpy-znp@{zigpy_znp.__version__}",
            "internal": {
                "creation_time": now.isoformat(timespec="seconds"),
                "zstack": {
                    "version": znp.version,
                },
            },
        },
        "coordinator_ieee": znp.network_info.ieee.serialize()[::-1].hex(),
        "pan_id": znp.network_info.pan_id.serialize()[::-1].hex(),
        "extended_pan_id": znp.network_info.extended_pan_id.serialize()[::-1].hex(),
        "nwk_update_id": znp.network_info.nwk_update_id,
        "security_level": znp.network_info.security_level,
        "channel": znp.network_info.channel,
        "channel_mask": list(znp.network_info.channels),
        "network_key": {
            "key": znp.network_info.network_key.serialize().hex(),
            "sequence_number": znp.network_info.network_key_seq,
            "frame_counter": await read_tc_frame_counter(znp),
        },
        "devices": devices,
    }

    if znp.version > 1.2:
        tclk_seed = await znp.nvram.osal_read(OsalNvIds.TCLK_SEED, item_type=t.Bytes)
        LOGGER.info("TCLK seed: %s", tclk_seed.hex())

        obj["stack_specific"] = {"zstack": {"tclk_seed": tclk_seed.hex()}}

    # Ensure our generated backup is valid
    validate_backup_json(obj)

    return obj


async def main(argv: list[str]) -> None:
    parser = setup_parser("Backup adapter network settings")
    parser.add_argument(
        "--output", "-o", type=ClosableFileType("w"), help="Output file", default="-"
    )
    args = parser.parse_args(argv)

    with args.output as f:
        znp = ZNP(ControllerApplication.SCHEMA({"device": {"path": args.serial}}))
        await znp.connect()

        backup_obj = await backup_network(znp)
        znp.close()

        f.write(json.dumps(backup_obj, indent=4))


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover
